
from betdaq.utils import clean_locals
from betdaq.enums import Boolean, WithdrawRepriceOption, OrderKillType


def create_order(SelectionId, Stake, Price, Polarity, ExpectedSelectionResetCount, ExpectedWithdrawalSequenceNumber,
                 CancelOnInRunning=Boolean.T, CancelIfSelectionReset=Boolean.T, ExpiresAt=None,
                 WithdrawalRepriceOption=WithdrawRepriceOption.Cancel, KillType=OrderKillType.FillOrKillDontCancel,
                 FillOrKillThreshold=0.0, PunterReferenceNumber=1):
    """
    Create an order to send to exchange.
    
    :param SelectionId: Id number of the selection on which the order is to be placed
    :type SelectionId: int
    :param Stake: Amount for which the order is to be placed for.
    :type Stake: float
    :param Price: Price at which order is to be placed.
    :type Price: float
    :param Polarity: side on which order is to be placed.
    :type Polarity: betdaq_py.enums.Polarity
    :param ExpectedSelectionResetCount: must match SelectionResetCount value in GetMarketInformation and GetPrices 
                                        to ensure state of the market before placing a bet. If not matching server
                                        bet will not be placed and error is raised.
    :type ExpectedSelectionResetCount: int
    :param ExpectedWithdrawalSequenceNumber: should match withdrawalSequenceNumber value in GetMarketInformation and 
                                             GetPrices. If not matching server then your bet WILL be accepted, but it 
                                             will be repriced.
    :type ExpectedWithdrawalSequenceNumber: int
    :param CancelOnInRunning: Cancel any unmatched orders when the market changes to an in-running market.
    :type CancelOnInRunning: betdaq_py.enums.Boolean
    :param CancelIfSelectionReset: Cancel any unmatched bets if the selection is reset. 
                                   This can occur when the Market is reset (eg a goal is scored).
    :type CancelIfSelectionReset: betdaq_py.enums.Boolean
    :param ExpiresAt: Specify a specific time for an order to expire, times in the past are instantly cancelled.
    :type ExpiresAt: datetime
    :param WithdrawalRepriceOption: Define what to do with the order in case of withdrawal, default to cancel.
    :type WithdrawalRepriceOption: betdaq_py.enums.WithdrawalRepriceOption
    :param KillType: whether to define order as a type of Fill/Kill order, default order will be limit order.
    :type KillType: betdaq_py.enums.OrderKillType
    :param FillOrKillThreshold: Lower limit for order to be partially filled. only required if KillType is 
                                FillOrKill or FillOrKillDontCancel
    :type FillOrKillThreshold: float
    :param PunterReferenceNumber: optional ID provided for customers own reference.
    :type PunterReferenceNumber: int
    :return: dictionary of all order information which can be sent to exchange.
    """
    return clean_locals(locals())


def update_order(BetId, DeltaStake, Price, ExpectedSelectionResetCount, ExpectedWithdrawalSequenceNumber,
                 CancelOnInRunning=None, CancelIfSelectionReset=None, SetToBeSPIfUnmatched=None):
    """
    
    :param BetId: ID of the bet to be updated.
    :type BetId: int
    :param DeltaStake: Amount to change the stake of the bet by.
    :type DeltaStake: float
    :param Price: Price at which to place bet at.
    :param ExpectedSelectionResetCount: must match SelectionResetCount value in GetMarketInformation and GetPrices 
                                        to ensure state of the market before placing a bet. If not matching server
                                        bet will not be placed and error is raised.
    :type ExpectedSelectionResetCount: int
    :param ExpectedWithdrawalSequenceNumber: should match withdrawalSequenceNumber value in GetMarketInformation and 
                                             GetPrices. If not matching server then your bet WILL be accepted, but it 
                                             will be repriced.
    :type ExpectedWithdrawalSequenceNumber: int
    :param CancelOnInRunning: Cancel any unmatched orders when the market changes to an in-running market.
    :type CancelOnInRunning: betdaq_py.enums.Boolean
    :param CancelIfSelectionReset: Cancel any unmatched bets if the selection is reset. 
                                   This can occur when the Market is reset (eg a goal is scored).
    :type CancelIfSelectionReset: betdaq_py.enums.Boolean
    :param SetToBeSPIfUnmatched: whether to set bet to SP when market turns in play if it is unmatched.
    :type SetToBeSPIfUnmatched: betdaq_py.enums.Boolean
    :return: dictionary of the order information to update on exchange.
    """
    return clean_locals(locals())




from betdaq.classifiers.marketdata import parse_classifier
from betdaq.utils import get_tag, get_attribs, listy_mc_list


def err_cancel_market(response):
    err_data = parse_classifier(response.get('_raw_elements', []), 'Order')
    response['Order'] = listy_mc_list(response.get('Order', [])) + err_data
    return response


def err_suspend_orders(response):
    err_data = parse_classifier(response.get('_raw_elements', []), 'Orders')
    response['Orders'] = listy_mc_list(response.get('Orders', [])) + err_data
    return response




from betdaq.classifiers.marketdata import (
    parse_classifier, parse_event_classfiers, parse_market_classifier, parse_selection_sequence_classifier,
    parse_selection_trades_classifier,
)
from betdaq.utils import get_attribs, get_tag, listy_mc_list


def err_selection_trades(response):
    err_data = parse_selection_trades_classifier(response.get('_raw_elements', []))
    response['SelectionTrades'] = listy_mc_list(response.get('SelectionTrades')) + err_data
    return response


def err_mkt_info(response):
    err_data = parse_market_classifier(response.get('_raw_elements', []))
    response['Markets'] = listy_mc_list(response.get('Markets', [])) + err_data.get('Markets', [])
    return response


def err_selection_changes(response):
    err_data = parse_selection_sequence_classifier(response.get('_raw_elements', []))
    response['Selections'] = listy_mc_list(response.get('Selections', [])) + err_data
    return response


def err_withdrawals(response):
    err_data = parse_classifier(response.get('_raw_elements', []), 'Withdrawals')
    response['Withdrawals'] = listy_mc_list(response.get('Withdrawals', [])) + err_data
    return response


def err_sport_markets(response):
    err_data = [
        {**get_attribs(mkt),
         **parse_market_classifier(mkt),
         **{'EventClassifiers': parse_event_classfiers(mkt)}} for mkt in response.get('_raw_elements', [])
    ]
    response['EventClassifiers'] = listy_mc_list(response.get('EventClassifiers', [])) + err_data
    return response


def err_sp_events(response):
    err_data = [
        {**get_attribs(child), **{'MarketTypeIds': {'MarketTypeId': sub_child.values()}
                                  for sub_child in child}} for child in response.get('_raw_elements', [])
    ]
    response['SPEnabledEvent'] = listy_mc_list(response.get('SPEnabledEvent', [])) + err_data
    return response


def err_prices(response):
    err_data = [
        {**get_attribs(mkt_prc),
         **{"Selections": [{**get_attribs(sel),
                            **{'_value_1': [{get_tag(prc): get_attribs(prc)} for prc in sel]}}
                           for sel in mkt_prc]}} for mkt_prc in response.get('_raw_elements', [])
    ]
    response['MarketPrices'] = [response['MarketPrices']] + err_data
    return response


def err_sports(response):
    err_data = [{**get_attribs(mkt), **parse_market_classifier(mkt)} for mkt in response.get('_raw_elements', [])]
    response['EventClassifiers'] = [response['EventClassifiers']] + err_data
    return response






from betdaq.resources.baseresources import BaseResource


class BlacklistInfo(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'ApiNamesAndTimes'
        attributes = {
            'ApiName': 'ApiName',
            'RemainingMS': 'RemainingMS',
        }


class Pulse(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = ''
        attributes = {
            'PerformedAt': 'PerformedAt',
            'HeartbeatAction': 'HeartbeatAction',
        }
        datetime_attributes = (
            'PerformedAt'
        )




import pytz
import datetime
from dateutil.parser import parse

from betdaq.exceptions import *
from betdaq.enums import ErrorMap


def get_tag(elem):
    return elem.tag.split('}')[-1]


def get_attribs(elem):
    return {k: v for k, v in elem.attrib.items()}


def listy_mc_list(maybe_list):
    if maybe_list:
        return maybe_list if isinstance(maybe_list, list) else [maybe_list]
    else:
        return []


price_side_map = {
    'ForSidePrices': 'back',
    'AgainstSidePrices': 'lay',
    1: 'back',
    2: 'lay',
}


def parse_time_str(time_str):
    return parse(time_str)


def make_tz_naive(date):
    if isinstance(date, str):
        try:
            date = parse_time_str(date).strftime('%Y-%m-%d %H:%M:%S.%f')
        except:
            pass
    if isinstance(date, datetime.datetime):
        date = date.astimezone(pytz.UTC).replace(tzinfo=None).strftime('%Y-%m-%d %H:%M:%S.%f')
    return date


def clean_locals(params):
    """
    Clean up locals dict, remove empty and self params.

    :param params: locals dicts from a function.
    :type params: dict
    :returns: cleaned locals dict to use as params for functions
    :rtype: dict
    """
    return dict((k, v) for k, v in params.items() if v is not None and k != 'self')


def check_status_code(response, codes=None):
    """Checks response status_code is in codes
    :param response: Requests response
    :param codes: List of accepted codes or callable
    :raises: StatusCodeError if code invalid
    """
    codes = codes or [0]
    response_code = response.get('ReturnStatus', {}).get('Code')
    if response_code not in codes:
        raise eval(ErrorMap(response_code).name)




import datetime

from betdaq.utils import clean_locals
from betdaq.enums import Boolean, PriceFormat
from betdaq.endpoints.baseendpoint import BaseEndpoint
from betdaq.utils import listy_mc_list
from betdaq.errorparsers.marketdata import (
    err_mkt_info, err_prices, err_selection_changes, err_selection_trades, err_sp_events, err_sport_markets,
    err_sports, err_withdrawals
)
from betdaq.resources.marketdataresources import (
    parse_sports, parse_deep_markets, parse_market, parse_selection_changes, parse_market_withdrawal,
    parse_market_prices, parse_ladder, parse_selection_trades,
)


class MarketData(BaseEndpoint):

    def get_sports(self, WantPlayMarkets=None):
        """
        Get list of sports and their IDs.
        
        :param WantPlayMarkets: whether to return play or real markets
        :return: 
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListTopLevelEvents', params, secure=False)
        data = self.process_response(response, date_time_sent, 'EventClassifiers', error_handler=err_sports)
        return [parse_sports(sport) for sport in listy_mc_list(data.get('data', []))] if data.get('data') else []

    def get_sport_markets(
            self, sport_ids, include_selections=True, WantDirectDescendentsOnly=Boolean.F.value, WantPlayMarkets=None
    ):
        """
        Get the tree of events and markets for given sports/events.
        
        :param sport_ids: list of sports for which to return events/markets
        :param include_selections: whether to include the selections in returned data.
        :param WantDirectDescendentsOnly: whether to return only direct descendents of the event.
        :param WantPlayMarkets: whether information about real or play markets should be returned
        :return: all markets for the given sport, with comp and event data flattened.
        """
        date_time_sent = datetime.datetime.utcnow()
        if include_selections:
            method = 'GetEventSubTreeWithSelections'
            params = self.client.readonly_types['%sRequest' % method](
                _value_1=[{'EventClassifierIds': s_id} for s_id in listy_mc_list(sport_ids)],
                WantPlayMarkets=WantPlayMarkets,
            )
        else:
            method = 'GetEventSubTreeNoSelections'
            params = self.client.readonly_types['%sRequest' % method](
                _value_1=[{'EventClassifierIds': s_id} for s_id in listy_mc_list(sport_ids)],
                WantDirectDescendentsOnly=WantDirectDescendentsOnly,
                WantPlayMarkets=WantPlayMarkets,
            )
        response = self.request(method, params, secure=False)
        data = self.process_response(response, date_time_sent, 'EventClassifiers', error_handler=err_sport_markets)
        return parse_deep_markets(listy_mc_list(data.get('data', []))) if data.get('data') else []

    def get_markets(self, market_ids):
        """
        Get detailed information about given market(s).
        
        :param market_ids: market id(s) to get data for.
        :return: market information for each market id provided.
        """
        date_time_sent = datetime.datetime.utcnow()
        params = self.client.readonly_types['GetMarketInformationRequest'](
            _value_1=[{'MarketIds': m_id} for m_id in listy_mc_list(market_ids)]
        )
        response = self.request('GetMarketInformation', params, secure=False)
        data = self.process_response(response, date_time_sent, 'Markets', error_handler=err_mkt_info)
        return parse_market(listy_mc_list(data.get('data', [])), {})if data.get('data') else []

    def get_selection_changes(self, SelectionSequenceNumber):
        """
        Poll to see if any selections have changed since the previous poll.
        
        :param SelectionSequenceNumber: sequence of the poll to check diffs from.
        :return: any changes to selections since the given sequence number.
        """
        date_time_sent = datetime.datetime.utcnow()
        params = clean_locals(locals())
        response = self.request('ListSelectionsChangedSince', params, secure=False)
        data = self.process_response(response, date_time_sent, 'Selections', error_handler=err_selection_changes)
        return [parse_selection_changes(chg) for chg in listy_mc_list(data.get('data', []))] if data.get('data') else []

    def get_market_withdrawals(self, MarketId):
        """
        Get the prices for a particular market.

        :param MarketId: ID of the market to check for withdrawals.
        :return: any withdrawals from the market.
        """
        date_time_sent = datetime.datetime.utcnow()
        params = clean_locals(locals())
        response = self.request('ListMarketWithdrawalHistory', params, secure=False)
        data = self.process_response(response, date_time_sent, 'Withdrawals', error_handler=err_withdrawals)
        return [parse_market_withdrawal(mkt) for mkt in listy_mc_list(data.get('data', []))] if data.get('data') else []

    def get_prices(self, market_ids, ThresholdAmount=5., NumberForPricesRequired=-1, NumberAgainstPricesRequired=-1,
                   WantMarketMatchedAmount=Boolean.T.value, WantSelectionsMatchedAmounts=Boolean.T.value,
                   WantSelectionMatchedDetails=Boolean.T.value):
        """
        Get prices by selection for all markets given in market_ids.
        
        :param market_ids: Markets to get prices for.
        :param ThresholdAmount: Minimum backers stake at a level for that level to be included.
        :param NumberForPricesRequired: Depth of the Back side of the book to cut off at, -1 returns all available.
        :param NumberAgainstPricesRequired: Depth of the Lay side of the book to cut off at, -1 returns all available.
        :param WantMarketMatchedAmount: whether or not the total amount matched on the market is returned.
        :param WantSelectionsMatchedAmounts: whether or not the total amount matched on each selection is returned. 
        :param WantSelectionMatchedDetails: whether or not details about the last match occurring on each selection.
        :return: Prices for each selection in each market.
        """
        date_time_sent = datetime.datetime.utcnow()
        params = self.client.readonly_types['GetPricesRequest'](
            _value_1=[{'MarketIds': e_id} for e_id in market_ids], ThresholdAmount=ThresholdAmount,
            NumberForPricesRequired=NumberForPricesRequired, NumberAgainstPricesRequired=NumberAgainstPricesRequired,
            WantMarketMatchedAmount=WantMarketMatchedAmount, WantSelectionsMatchedAmounts=WantSelectionsMatchedAmounts,
            WantSelectionMatchedDetails=WantSelectionMatchedDetails,
        )
        response = self.request('GetPrices', params, secure=False)
        data = self.process_response(response, date_time_sent, 'MarketPrices', error_handler=err_prices)
        return [parse_market_prices(mkt) for mkt in listy_mc_list(data.get('data', []))] if data.get('data') else []

    def get_odds_ladder(self, PriceFormat=PriceFormat.Decimal.value):
        """
        Get current odds ladder.
        
        :param PriceFormat: what odds type to return. 
        :return: odds ladder
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('GetOddsLadder', params, secure=False)
        data = self.process_response(response, date_time_sent, 'Ladder')
        return parse_ladder(data.get('data', [])) if data.get('data') else []

    def get_markets_with_sp(self):
        """
        Get information defining which markets are enabled for starting-price orders.
        
        :return: information on which markets have SP.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('GetSPEnabledMarketsInformation', {}, secure=False)
        data = self.process_response(response, date_time_sent, 'SPEnabledEvent', error_handler=err_sp_events)
        return data.get('data', [])

    def get_selection_sequence_number(self):
        """
        Get the current maximum selectionSequenceNumber.
        
        :return: max selection sequence number.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('GetCurrentSelectionSequenceNumber', {}, secure=False)
        data = self.process_response(response, date_time_sent, 'SelectionSequenceNumber')
        return data.get('data')

    def get_selection_trades(self, selection_info, currency='EUR'):
        """
        Get the history of trades on the selection(s) specified.
        
        :param selection_info: list of dicts of selection information.
        :param currency: 3 letter code for currency to return trade info in.
        :return: trades on any selections provided from the tradeId specified.
        """
        params = self.client.readonly_types['ListSelectionTradesRequest'](
            _value_1=[{'selectionRequests': {'selectionId':  s_i.get('selectionId'),
                                             'fromTradeId': s_i.get('fromTradeId')}} for s_i in selection_info],
            currency=currency,
        )
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListSelectionTrades', params, secure=False)
        data = self.process_response(response, date_time_sent, 'SelectionTrades', error_handler=err_selection_trades)
        return [parse_selection_trades(trd) for trd in listy_mc_list(data.get('data', []))] if data.get('data') else []




from betdaq.utils import get_attribs, get_tag


def parse_selection_trades_classifier(elem):
    return [
        {**get_attribs(e), **{'_value_1': [{'TradeItems': t} for t in parse_classifier(e, 'TradeItems')]}}
        for e in elem if get_tag(e) == 'SelectionTrades'
    ]


def parse_selection_sequence_classifier(elem):
    return [
        {**get_attribs(e), **{'_value_1': [
            {'SettlementInformation': s_info} for s_info in parse_classifier(e, 'SettlementInformation')]}}
        for e in elem if get_tag(e) == 'Selections']


def parse_event_classfiers(elem):
    return [
        {**get_attribs(e), **parse_market_classifier(e),
         **{'EventClassifiers': [
             {**get_attribs(t), **parse_market_classifier(t),
              **{'EventClassifiers': [
                  {**get_attribs(c), **parse_market_classifier(c),
                   **{'EventClassifiers': [{**get_attribs(g), **parse_market_classifier(g)}
                                           for g in c if get_tag(g) == 'EventClassifiers']}}
                  for c in t if get_tag(c) == 'EventClassifiers']}}
             for t in e if get_tag(t) == 'EventClassifiers']}}
        for e in elem if get_tag(e) == 'EventClassifiers']


def parse_market_classifier(elem):
    return {'Markets': [{**get_attribs(e), **{'Selections': parse_classifier(e, 'Selections')}}
                        for e in elem if get_tag(e) == 'Markets']}


def parse_classifier(elem, name):
    return [get_attribs(e) for e in elem if get_tag(e) == name]




import datetime

from betdaq import resources
from betdaq.enums import HeartbeatAction
from betdaq.utils import clean_locals
from betdaq.endpoints.baseendpoint import BaseEndpoint


class Trading(BaseEndpoint):

    def list_blacklist_information(self):
        """
        Lists the black-list status for the punter.
        
        :return: list of every API from which the Punter is currentltly black-listed along with the remaining time 
                 (in milli-seconds) of the black-list period for that API. 
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListBlacklistInformation', {}, secure=True)
        return self.process_response(response, resources.BlacklistInfo, date_time_sent,  datetime.datetime.utcnow())

    def suspend_from_trading(self):
        """
        Suspend any of your orders from being matched. For emergency use only.
        
        :return: response of request success.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('SuspendFromTrading', {}, secure=True)
        return self.process_response(response, resources.EmptyResource, date_time_sent,  datetime.datetime.utcnow())

    def unsuspend_from_trading(self):
        """
        Reverse the suspension of trading. All active orders must be cancelled or suspended before you can unsuspend.

        :return: response of request success.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('UnsuspendFromTrading', {}, secure=True)
        return self.process_response(response, resources.EmptyResource, date_time_sent, datetime.datetime.utcnow())

    def register_heartbeat(self, HeartbeatAction=HeartbeatAction.CancelOrders.value, ThresholdMs=6000):
        """
        Register the Punter as requiring a Heartbeat. Must send a Pulse < every ThresholdMs to stay alive.
        
        :param HeartbeatAction: The action that should be taken if a Pulse is not received within the threshold. 
        :type HeartbeatAction: betdaq_py.enums.HeartbeatAction
        :param ThresholdMs: The maximum period (in milli-seconds) that can elapse between Pulse API calls being 
                            received before the system takes the relevant action. 
        :return: response of request success.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('RegisterHeartbeat', params, secure=True)
        return self.process_response(response, resources.EmptyResource, date_time_sent, datetime.datetime.utcnow())

    def deregister_heartbeat(self):
        """
        Deregister the Punter as requiring a Heartbeat.

        :return: response of request success.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('DeregisterHeartbeat', params, secure=True)
        return self.process_response(response, resources.EmptyResource, date_time_sent, datetime.datetime.utcnow())

    def change_hearbeat(self, HeartbeatAction=HeartbeatAction.CancelOrders.value, ThresholdMs=6000):
        """
        Update the parameter of Heartbeat.

        :param HeartbeatAction: The action that should be taken if a Pulse is not received within the threshold. 
        :type HeartbeatAction: betdaq_py.enums.HeartbeatAction
        :param ThresholdMs: The maximum period (in milli-seconds) that can elapse between Pulse API calls being 
                            received before the system takes the relevant action. 
        :return: response of request success.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ChangeHeartbeatRegistration', params, secure=True)
        return self.process_response(response, resources.EmptyResource, date_time_sent, datetime.datetime.utcnow())

    def send_pulse(self):
        """
        Notify the system that the application is still active and still has connectivity.
        
        :return: Time of pulse performance.
        """
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('Pulse', {}, secure=True)
        return self.process_response(response, resources.Pulse, date_time_sent, datetime.datetime.utcnow())




from enum import Enum

class HeartbeatAction(Enum):
    """
    Action to be performed when a threshold period has expired without a Pulse having been received.
    
    :var CancelOrders: cancel all unmatched orders.
    :var SuspendOrders: suspend all unmatched orders.
    :var SuspendPunter: suspend punter.
    """
    CancelOrders = 1
    SuspendOrders = 2
    SuspendPunter = 3


class OrderKillType(Enum):
    """
    Define how an order is handled when sent to exchange.
    
    :var FillAndKill: After the initial attempt is made to match this order any unmatched portion of the order is 
    immediately cancelled.
    :var FillOrKill: On the initial attempt to match this order if it is not possible to match a specified amount of 
    the order then none of the order will be matched. If it had been possible to match at least the specified amount 
    then the amount that can be matched will be matched and the remaining unmatched amount will be cancelled.
    :var FillOrKillDontCancel: On the initial attempt to match this order if it is not possible to match a specified 
    mount of the order then none of the order will be matched. If it had been possible to match at least the specified 
    amount then the amount that can be matched will be matched and the remaining unmatched amount will be not be 
    cancelled but left as an unmatched order.
    :var SPIfUnmatched: Same as Normal but any unmatched portion of the order is to be matched at SP when the market 
    is turned in-running (or completed).
    """
    Normal = 1
    FillAndKill = 2
    FillOrKill = 3
    FillOrKillDontCancel = 4
    SPIfUnmatched = 5


class MarketStatus(Enum):
    """
    The current status of a market.
    
    :var INACTIVEW: The Market is not active and has never had any Orders issued against it.
    :var ACTIVE: The Market it active (that is, Orders can be issued against it).
    :var SUSPENDED: The Market is not currently active but it has not yet been completed.
    :var CLOSED: The Market is completed. No further Orders can be issued against the Market but the result of the 
    Market is either not yet known or has not yet been entered.
    :var SETTLED: The Market has been fully settled.
    :var VOIDED: The Market has been voided. All matched Orders in this Market have also been voided.
    """
    INACTIVE = 1
    ACTIVE = 2
    SUSPENDED = 3
    CLOSED = 4
    SETTLED = 6
    VOIDED = 7


class MarketType(Enum):
    """
    The type of a market.
    """
    Win = 1
    Place = 2
    MatchOdds = 3
    OverUnder = 4
    AsianHandicap = 10
    TwoBall = 11
    ThreeBall = 12
    Unspecified = 13
    MatchMarket = 14
    SetMarket = 15
    Moneyline = 16
    Total = 17
    Handicap = 18
    EachWayNonHandicap = 19
    EachWayHandicap = 20
    EachWayTournament = 21
    RunningBall = 22
    MatchBetting = 23
    MatchBettingInclDraw = 24
    CorrectScore = 25
    HalfTimeFullTime = 26
    TotalGoals = 27
    GoalsScored = 28
    Corners = 29
    OddsOrEvens = 30
    HalfTimeResult = 31
    HalfTimeScore = 32
    MatchOddsExtraTime = 33
    CorrectScoreExtraTime = 34
    OverUnderExtraTime = 35
    ToQualify = 36
    DrawNoBet = 37
    HalftimeAsianHcp = 39
    HalftimeOverUnder = 40
    NextGoal = 41
    FirstGoalscorer = 42
    LastGoalscorer = 43
    PlayerToScore = 44
    FirstHalfHandicap = 45
    FirstHalfTotal = 46
    SetBetting = 47
    GroupBetting = 48
    MatchplaySingle = 49
    MatchplayFourball = 50
    MatchplayFoursome = 51
    TiedMatch = 52
    TopBatsman = 53
    InningsRuns = 54
    TotalTries = 55
    TotalPoints = 56
    FrameBetting = 57
    ToScoreFirst = 58
    ToScoreLast = 59
    FirstScoringPlay = 60
    LastScoringPlay = 61
    HighestScoringQtr = 62
    RunLine = 63
    RoundBetting = 64
    LineBetting = 65


class OrderActionType(Enum):
    """
    The type of an order history audit record.
    """
    Placed = 1
    ExplicitlyUpdated = 2
    Matched = 3
    CancelledExplicitly = 4
    CancelledByReset = 5
    CancelledOnInRunning = 6
    Expired = 7
    MatchedPortionRepricedByR4 = 8
    UnmatchedPortionRepricedByR4 = 9
    UnmatchedPortionCancelledByWithdrawal = 10
    Voided = 11
    Settled = 12
    Suspended = 13
    Unsuspended = 14
    ExpiredByMatching = 15
    Unsettled = 16
    Unmatched = 17
    MatchedPortionRepriced = 18
    CreatedFromLightweightPrice = 19
    CancelledOnComplete = 20


class OrderStatus(Enum):
    """
    The status of an order.
    
    :var Unmatched: The order is active and has some amount available for matching (the order may be partially matched).
    :var Matched: The order has not been settled and it does not have any unmatched amount. Either the order was fully 
    matched or it was partially matched and then cancelled.
    :var Cancelled: This order has been cancelled and at least some of the order was unmatched at the time of 
    expiration.
    :var Settled: The order has been settled.
    :var Voided: The order has been voided.
    :var Suspended: At least some of this order is unmatched but the order is suspended and is not available 
    for matching.
    """
    Unmatched = 1
    Matched = 2
    Cancelled = 3
    Settled = 4
    Voided = 5
    Suspended = 6


class Polarity(Enum):
    """
    The side which an order is being sent for.
    """
    back = 1
    lay = 2


class PostingCategory(Enum):
    """
    The categories of a posting.
    
    :var Settlement: This posting resulted from the settlement, unsettlement or resettlement of a specific order.
    :var Commission: This posting resulting from the charging of commission on market settlement, unsettlement or 
    resettlement.
    :var Other: This posting resulted from any other cause (for example, a lodgement or withdrawal). 
    """
    Settlement = 1
    Commission = 2
    Other = 3


class PunterCommissionBasis(Enum):
    """
    The basis on which commission for the punter is to be calculated.
    """
    NetMarketWinnings = 1
    MatchedForSideStakeExcludingPush = 2
    MatchedForSideStakeIncludingPush = 3
    MatchedAgainstSideStakeExclusingPush = 4
    MatchedAgainstSideStakeIncludingPush = 5
    MatchedRiskedStakeExcludingPush = 6
    MatchedRiskedStakeIncludingPush = 7


class PriceFormat(Enum):
    """
    The format of a price.
    
    :var Decimal: The price is expressed in decimal format – in particular the price is the decimal representation of 
    the payout for 1 currency unit stake.
    :var Fractional: The price is expressed in fractional format – in particular the fraction is the winnings 
    for 1 currency unit stake.
    :var American: The price is expressed in American format – in particular if the price is greater than 0 it means 
    the amount of winnings for 100 currency unit stake whereas if the price is less than zero it means the 
    amount of currency unit that needs to be staked to win 100 currency units.
    """
    Decimal = 1
    Fractional = 2
    American = 3


class SelectionStatus(Enum):
    """
    The current status of a Selection.
    
    :var Inactive: The Selection is not active and has never had any Orders issued against it.
    :var Active: The Selection is active (that is, Orders can be issued against it).
    :var Suspended: Orders can not currently be placed on this Selection.
    :var Withdrawn: The Entrant explicitly referenced by the Selection has withdrawn from the Event and so Orders 
    can no longer be placed on this Selection.
    :var BallotedOut: The Entrant explicitly referenced by the Selection has been balloted-out.
    :var Voided: Orders can no longer be placed on this Selection and Orders previously placed for or against 
    the Selection have been voided.
    :var Completed: The Selection is completed. No further Orders can be issued against the Selection but the 
    result of the Selection is either not yet known or has not yet been entered.
    :var Settled: This Selection has already been settled. Individual selections can be settled in advance of 
    other selections in the market being settled (early settlement).
    """
    Inactive = 1
    Active = 2
    Suspended = 3
    Withdrawn = 4
    BallotedOut = 9
    Voided = 5
    Completed = 6
    Settled = 8


class WithdrawRepriceOption(Enum):
    """
    the action to take on a specific Order if a withdrawal occurs on the Market that could cause a 
    Rule-4 deduction factor to be applied to the Order (this option only controls what happens to the 
    unmatched parts of Orders, those parts that have already been matched will have the rule-4 deduction applied 
    regardless of the value of this option). 
    
    :var Reprice: Reprice the unmatched parts of the Order. It is anticipated that this would be the 
    usual option specified by Layers.
    :var Cancel: Cancel the unmatched parts of the Order.
    :var DontReprice: Do not reprice the unmatched parts of the Order.
    """
    Reprice = 1
    Cancel = 2
    DontReprice = 3


class Boolean(Enum):
    """
    Booleans must be send as lower case strings to soap requests.
    """
    T = 'true'
    F = 'false'


class SportID(Enum):
    """
    Hard Code sports ID as endpoint is currently broken
    """
    HorseRacing = 100004
    Soccer = 100003
    GreyhoundRacing = 100008
    Golf = 100006
    Cricket = 100007
    Tennis = 100005
    RugbyUnion = 100010
    RugbyLeague = 100009
    AmericanFootball = 100014
    Basketball = 100017
    FormulaOne = 100015
    Baseball = 100016
    Boxing = 100020
    AustralianRules = 338839
    GAA = 100013
    RacingBloggerSpecials = 487047
    Politics = 481576
    VirtualSports = 1457259


class ErrorMap(Enum):
    """
    Map ReturnStatus Codes to their Specific Errors.
    """
    ResourceError = 1
    BetdaqSystemError = 2
    EventClassifierDoesNotExist = 5
    MarketDoesNotExist = 8
    SelectionDoesNotExist = 11
    MarketNotActive = 15
    MarketNeitherSuspendedNorActive = 16
    SelectionNotActive = 17
    InsufficientVirtualPunterFunds = 19
    OrderDoesNotExist = 21
    NoUnmatchedAmount = 22
    ResetHasOccurred = 114
    OrderAlreadySuspended = 127
    TradingCurrentlySuspended = 128
    InvalidOdds = 131
    WithdrawalSequenceNumberIsInvalid = 136
    MaximumInputRecordsExceeded = 137
    PunterSuspended = 208
    PunterProhibitedFromPlacingOrders = 240
    InsufficientPunterFunds = 241
    OrderAPIInProgress = 271
    PunterOrderMismatch = 274
    MarketNotEnabledForMultiples = 281
    MultipleLayerParameterAlreadyExists = 285
    LevelsRequestedExceedsMaximum = 288
    NoMultipleOfferAvailable = 289
    InRunningDelayInEffect = 293
    MultipleSelectionsUnderSameEvent = 295
    MultipleSelectionsWithSameName = 296
    DuplicateOrderSpecified = 299
    OrderNotSuspended = 301
    PunterIsSuspendedFromTrading = 302
    PunterHasActiveOrders = 303
    PunterNotSuspendedFromTrading = 304
    ExpiryTimeInThePast = 305
    NoChangeSpecified = 306
    SoapHeaderNotSupplied = 307
    IncorrectVersionNumber = 308
    NoUsernameSpecified = 309
    InvalidParameters = 310
    NoPasswordSpecified = 311
    MultipleCombinationExclusionAlreadyExists = 312
    MultipleCombinationExclusionDoesNotExist = 313
    InvalidPassword = 405
    PunterIsBlacklisted = 406
    PunterNotRegisteredAsMultipleLayer = 425
    PunterAlreadyRegisteredForHeartbeat = 462
    PunterNotRegisteredForHeartbeat = 463
    ThresholdSpecifiedTooSmall = 473
    UnmatchedOrderCouldResult = 477
    PunterNotAuthorisedForAPI = 533
    MarketIsForRealMoney = 597
    MarketIsForPlayMoney = 598
    CannotChangeToSPIfUnmatched = 892
    InvalidUsername = 1000




import datetime

from betdaq.utils import clean_locals
from betdaq.endpoints.baseendpoint import BaseEndpoint
from betdaq.resources.accountresources import parse_account_balance, parse_account_postings


class Account(BaseEndpoint):

    def get_account_balances(self):
        """
        Get summary of current balances.

        :return: account information for logged in user.
        :rtype: dict
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('GetAccountBalances', params, secure=True)
        data = self.process_response(response, date_time_sent, None)
        return parse_account_balance(data.get('data', {})) if data.get('data') else {}

    def get_account_transactions(self, StartTime, EndTime):
        """
        Get account transactions between two given date and times.

        :param StartTime: earlier time to include transactions from.
        :type StartTime: Timestamp
        :param EndTime: latest time to include transactions until.
        :type EndTime: Timestamp
        :return: account transactions over the specified period.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListAccountPostings', params, secure=True)
        data = self.process_response(response, date_time_sent, None)
        return parse_account_postings(data.get('data', {})) if data.get('data') else {}

    def get_account_transactions_by_id(self, TransactionId):
        """
        Get account transactions with transactionId greater than that specified.

        :param TransactionId: lower cutoff for transactionIds to include.
        :type TransactionId: int
        :return: account transactions greater than specified transactionId.
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ListAccountPostingsById', params, secure=True)
        data = self.process_response(response, date_time_sent, None)
        return parse_account_postings(data.get('data', {})) if data.get('data') else {}

    def change_account_password(self, Password):
        """
        Change the password for the logged in user.

        :param Password: new password to be used for logged in user.
        :type Password: str
        :return: None
        """
        params = clean_locals(locals())
        date_time_sent = datetime.datetime.utcnow()
        response = self.request('ChangePassword', params, secure=True)
        data = self.process_response(response, date_time_sent, None)
        return data.get('data', {})




import zeep


class BaseClient:
    def __init__(self, username, password):
        self.username = username
        self.password = password
        self.wsdl_file = 'http://api.betdaq.com/v2.0/API.wsdl'
        self.readonly_types = None
        self.secure_types = None
        self.secure_client, self.readonly_client = self.initialise_wsdl()
        self.initialise_type_factories()

    def initialise_wsdl(self):
        """
        Inspect the WSDL document and generates the corresponding code to use the services and types in the document.
        
        :return: client with available endpoints mapped.
        :rtype: zeep.Client
        """
        secure_client = zeep.Client(
            wsdl=self.wsdl_file, service_name='SecureService', port_name='SecureService', strict=False
        )
        secure_client.set_default_soapheaders({'ExternalApiHeader': self.external_headers})
        readonly_client = zeep.Client(
            wsdl=self.wsdl_file, service_name='ReadOnlyService', port_name='ReadOnlyService', strict=False
        )
        readonly_client.set_default_soapheaders({'ExternalApiHeader': self.external_headers})
        return secure_client, readonly_client

    def initialise_type_factories(self):
        self.readonly_types = self.readonly_client.type_factory('ns0')
        self.secure_types = self.secure_client.type_factory('ns0')

    @property
    def external_headers(self):
        return {"version": 2.0,
                "languageCode": 'en',
                "username": self.username,
                "password": self.password,
                "applicationIdentifier": None}
